LIBRARY "freeglut.dll"
EXPORTS
__glutCreateMenuWithExit@8
__glutCreateWindowWithExit@8
__glutInitWithExit@12
glutAddMenuEntry@8
glutAddSubMenu@8
glutAttachMenu@4
glutBitmapCharacter@8
glutBitmapHeight@4
glutBitmapLength@8
glutBitmapString@8
glutBitmapWidth@8
glutButtonBoxFunc@4
glutChangeToMenuEntry@12
glutChangeToSubMenu@12
glutCloseFunc@4
glutCopyColormap@4
glutCreateMenu@4
glutCreateSubWindow@20
glutCreateWindow@4
glutDestroyMenu@4
glutDestroyWindow@4
glutDetachMenu@4
glutDeviceGet@4
glutDialsFunc@4
glutDisplayFunc@4
glutEnterGameMode@0
glutEntryFunc@4
glutEstablishOverlay@0
glutExit@0
glutExtensionSupported@4
glutForceJoystickFunc@0
glutFullScreen@0
glutFullScreenToggle@0
glutGameModeGet@4
glutGameModeString@4
glutGet@4
glutGetColor@8
glutGetMenu@0
glutGetMenuData@0
glutGetModeValues@8
glutGetModifiers@0
glutGetProcAddress@4
glutGetWindow@0
glutGetWindowData@0
glutHideOverlay@0
glutHideWindow@0
glutIconifyWindow@0
glutIdleFunc@4
glutIgnoreKeyRepeat@4
glutInit@8
glutInitContextFlags@4
glutInitContextProfile@4
glutInitContextVersion@8
glutInitDisplayMode@4
glutInitDisplayString@4
glutInitErrorFunc@4
glutInitWarningFunc@4
glutInitWindowPosition@8
glutInitWindowSize@8
glutJoystickFunc@8
glutKeyboardFunc@4
glutKeyboardUpFunc@4
glutLayerGet@4
glutLeaveFullScreen@0
glutLeaveGameMode@0
glutLeaveMainLoop@0
glutMainLoop@0
glutMainLoopEvent@0
glutMenuDestroyFunc@4
glutMenuStateFunc@4
glutMenuStatusFunc@4
glutMotionFunc@4
glutMouseFunc@4
glutMouseWheelFunc@4
glutMultiButtonFunc@4
glutMultiEntryFunc@4
glutMultiMotionFunc@4
glutMultiPassiveFunc@4
glutOverlayDisplayFunc@4
glutPassiveMotionFunc@4
glutPopWindow@0
glutPositionWindow@8
glutPostOverlayRedisplay@0
glutPostRedisplay@0
glutPostWindowOverlayRedisplay@4
glutPostWindowRedisplay@4
glutPushWindow@0
glutRemoveMenuItem@4
glutRemoveOverlay@0
glutReportErrors@0
glutReshapeFunc@4
glutReshapeWindow@8
glutSetColor@16
glutSetCursor@4
glutSetIconTitle@4
glutSetKeyRepeat@4
glutSetMenu@4
glutSetMenuData@4
glutSetOption@8
glutSetWindow@4
glutSetWindowData@4
glutSetWindowTitle@4
glutSetupVideoResizing@0
glutShowOverlay@0
glutShowWindow@0
glutSolidCone@24
glutSolidCube@8
glutSolidCylinder@24
glutSolidDodecahedron@0
glutSolidIcosahedron@0
glutSolidOctahedron@0
glutSolidRhombicDodecahedron@0
glutSolidSierpinskiSponge@16
glutSolidSphere@16
glutSolidTeapot@8
glutSolidTetrahedron@0
glutSolidTorus@24
glutSpaceballButtonFunc@4
glutSpaceballMotionFunc@4
glutSpaceballRotateFunc@4
glutSpecialFunc@4
glutSpecialUpFunc@4
glutStopVideoResizing@0
glutStrokeCharacter@8
glutStrokeHeight@4
glutStrokeLength@8
glutStrokeString@8
glutStrokeWidth@8
glutSwapBuffers@0
glutTabletButtonFunc@4
glutTabletMotionFunc@4
glutTimerFunc@12
glutUseLayer@4
glutVideoPan@16
glutVideoResize@16
glutVideoResizeGet@4
glutVisibilityFunc@4
glutWMCloseFunc@4
glutWarpPointer@8
glutWindowStatusFunc@4
glutWireCone@24
glutWireCube@8
glutWireCylinder@24
glutWireDodecahedron@0
glutWireIcosahedron@0
glutWireOctahedron@0
glutWireRhombicDodecahedron@0
glutWireSierpinskiSponge@16
glutWireSphere@16
glutWireTeapot@8
glutWireTetrahedron@0
glutWireTorus@24
